/////////////////////////////////////////////////////////////////////////////
//
// Windows MFC Glk Libraries
//
// GlkSoundMOD
// Glk MOD module sounds
//
/////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "GlkSoundMOD.h"
#include "Resource.h"

extern "C"
{
#include "gi_blorb.h"
}

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// Class for MOD sound loader
/////////////////////////////////////////////////////////////////////////////

// Get file prefix for sounds supported by this loader
LPCTSTR CWinGlkMODSoundLoader::GetFilePrefix(void)
{
	return "mus";
}

// Get the file extensions for sounds supported for this loader
int CWinGlkMODSoundLoader::GetNumberFileExtensions(void)
{
	return 1;
}

LPCTSTR CWinGlkMODSoundLoader::GetFileExtension(int iExtIndex)
{
	return "mod";
}

// Get the identifier for sounds supported for this loader
glui32 CWinGlkMODSoundLoader::GetIdentifier(void)
{
	return giblorb_make_id('M','O','D',' ');
}

// Get a sound object
CWinGlkSound* CWinGlkMODSoundLoader::GetSound(LPCTSTR pszFileName)
{
	CWinGlkMODSound::InitMODPlug();
	return new CWinGlkMODSound(pszFileName);
}

CWinGlkSound* CWinGlkMODSoundLoader::GetSound(BYTE* pData, int iLength)
{
	CWinGlkMODSound::InitMODPlug();
	return new CWinGlkMODSound(pData,iLength);
}

// Clean up after sound playback has been stopped
void CWinGlkMODSoundLoader::SoundStopped(void)
{
	CWinGlkMODSound::CloseMODPlug();
}

/////////////////////////////////////////////////////////////////////////////
// Class for MOD sounds
/////////////////////////////////////////////////////////////////////////////

CWinGlkMODSound::CWinGlkMODSound(BYTE* pData, int iLength) : CWinGlkSound(pData,iLength)
{
	m_Duration = 0;
}

CWinGlkMODSound::CWinGlkMODSound(LPCTSTR pszFileName) : CWinGlkSound(pszFileName)
{
	m_Duration = 0;
}

CWinGlkMODSound::~CWinGlkMODSound()
{
	RemoveFromList();
}

bool CWinGlkMODSound::Play(int iRepeat, int iVolume)
{
	if (m_pData == NULL)
		return false;
	if (IsMODPlugLoaded() == false)
		return false;

	// Only one MOD can be playing at any given time, so stop
	// any currently playing MODs
	CDSoundEngine& Engine = CDSoundEngine::GetSoundEngine();
	Engine.StopSounds(GetType());

	// Load the song into MODPlug
	if (m_pMPPAPI->LoadSong(m_pData,m_iLength) != MPPERR_NOERROR)
		return false;

	// Set the number of repeats and the duration
	if (iRepeat > 0)
	{
		m_pMPPAPI->SetRepeatCount(iRepeat-1);
		m_Duration = m_pMPPAPI->GetSongLength() * iRepeat * 1000;
	}
	else
	{
		m_pMPPAPI->SetRepeatCount(-1);
		m_Duration = -1;
	}

	// Create a buffer
	WAVEFORMATEX& Format = Engine.GetPrimaryFormat();
	if (CreateBuffer(Format.nChannels,Format.nSamplesPerSec,Format.wBitsPerSample) == false)
		return false;

	// Fill the buffer with sample data
	if (FillBuffer(GetBufferSize()) == false)
		return false;

	// Set the volume for the buffer
	SetVolume(iVolume);

	// Start the buffer playing
	return PlayBuffer();
}

bool CWinGlkMODSound::IsPlaying(void)
{
	return IsBufferPlaying();
}

void CWinGlkMODSound::SetVolume(int iVolume)
{
	// The SetVolume() call to DirectSound requires a volume
	// in 100ths of a decibel.
	SetBufferVolume((LONG)DecibelVolume(iVolume) * 100L);
}

void CWinGlkMODSound::WriteSampleData(unsigned char* pSample, int iSampleLen)
{
	WAVEFORMATEX& Format = CDSoundEngine::GetSoundEngine().GetPrimaryFormat();

	// Render audio
	int iBytes = m_pMPPAPI->Render(pSample,iSampleLen) * Format.nBlockAlign;

	// Has the MOD been completely rendered?
	if (iBytes < iSampleLen)
	{
		for (int i = iBytes; i < iSampleLen; i++)
		{
			pSample[i] = (unsigned char)
				((Format.wBitsPerSample == 8) ? 0x80 : 0x00);
		}
	}
}

// Check if the sound has finished playing
bool CWinGlkMODSound::IsSoundOver(DWORD Tick)
{
	if (m_Playing == false)
		return true;

	// Check if sound is playing forever
	if (m_Duration < 0)
		return false;
	return (Tick > m_StartTime + m_Duration);
}

// Get a type identifier for the sound
int CWinGlkMODSound::GetType(void)
{
	return (int)'M';
}

/////////////////////////////////////////////////////////////////////////////
// MODPlug functions and data
/////////////////////////////////////////////////////////////////////////////

HINSTANCE CWinGlkMODSound::m_hMODPlugDLL = 0;
IModMixer* CWinGlkMODSound::m_pMPPAPI = NULL;
bool CWinGlkMODSound::m_bShowError = true;

// Initialize the MODPlug DLL
void CWinGlkMODSound::InitMODPlug(void)
{
	if (IsMODPlugLoaded())
		return;

	// Load the DLL
	m_hMODPlugDLL = AfxLoadLibrary("mppsdk.dll");
	if (m_hMODPlugDLL != 0)
	{
		// Get the exported function to access MODPlug
		MPP_GETMODAPIFUNC pGetAPI = (MPP_GETMODAPIFUNC)
			::GetProcAddress(m_hMODPlugDLL,MPP_GETMODAPIFUNCNAME);
		if (pGetAPI != NULL)
		{
			// Get the IModMixer interface
			pGetAPI(&m_pMPPAPI);
			if (m_pMPPAPI != NULL)
			{
				// Check the version number
				if (m_pMPPAPI->GetVersion() >= MPPAPI_VERSION)
				{
					// Set up the player
					WAVEFORMATEX& Format = CDSoundEngine::GetSoundEngine().
						GetPrimaryFormat();
					MPPERR Result = m_pMPPAPI->SetWaveFormat(Format.nSamplesPerSec,
						Format.nChannels,Format.wBitsPerSample);
					if (Result == MPPERR_NOERROR)
					{
						// Set the options for MODPlug
						Result = m_pMPPAPI->SetMixerOptions(
							MPPMIX_SURROUND|MPPMIX_HIGHQUALITY|MPPMIX_BASSEXPANSION);
						if (Result == MPPERR_NOERROR)
						{
							// Set the amplification level to maximum
							m_pMPPAPI->SetPreampLevel(400);
							return;
						}
					}
				}
			}
		}
	}

	// Show an error if MODPlug failed
	if (m_bShowError)
	{
		CString noModPlug;
		noModPlug.LoadString(IDS_NO_MODPLUG);

		::MessageBox(AfxGetMainWnd()->GetSafeHwnd(),
			noModPlug,"Windows Glk",MB_ICONWARNING|MB_OK);
		m_bShowError = false;
	}

	// If loading MODPlug has failed, clean up
	CloseMODPlug();
}

// Close the MODPlug DLL
void CWinGlkMODSound::CloseMODPlug(void)
{
	if (m_pMPPAPI != NULL)
	{
		m_pMPPAPI->Release();
		m_pMPPAPI = NULL;
	}

	if (m_hMODPlugDLL != 0)
	{
		AfxFreeLibrary(m_hMODPlugDLL);
		m_hMODPlugDLL = 0;
	}
}

// Check if MODPlug is loaded
bool CWinGlkMODSound::IsMODPlugLoaded(void)
{
	return (m_pMPPAPI != NULL);
}
